import time
import datetime
import pygiftgrab as pgg

use_numpy = True
try:
    import numpy as np
except ImportError:
    use_numpy = False


class FrameRateTimer(pgg.IObserver):
    """Descendant of GIFT-Grab's `Observer`, which
    will listen to `Observable`s for some time and
    when asked, will report whether data has been
    sent at the specified frame rate.
    """

    def __init__(self, frame_rate, init_time = 0):
        super(FrameRateTimer, self).__init__()
        self._frame_rate = frame_rate
        # all data for this initial period (in sec) will be discarded
        self._init_time = init_time
        if use_numpy:
            self._timestamps = np.array([], dtype='datetime64[us]')
        else:
            self._timestamps = []

    def update(self, frame):
        if use_numpy:
            self._timestamps = np.append(self._timestamps,
                np.datetime64(datetime.datetime.now()))
        else:
            self._timestamps.append(datetime.datetime.now())

    def __bool__(self):
        """Check if updates have been in time intervals
        in line with defined frame rate, also resetting
        all saved timestamps, i.e. ready for next round.
        """
        global use_numpy
        n_init_items = int(self._init_time * self._frame_rate)
        if use_numpy:
            if n_init_items >= self._timestamps.size:
                raise IndexError('Not enough data collected')
        else:
            if n_init_items >= len(self._timestamps):
                raise IndexError('Not enough data collected')
        timestamps = self._timestamps[n_init_items:]

        if use_numpy:
            intervals = timestamps[1:] - timestamps[0]
            frame_rates = np.array(
                [(i + 1) / (interval / np.timedelta64(1, 's'))
                    for i, interval in enumerate(intervals)]
            )
            return np.min(frame_rates) >= self._frame_rate
        else:
            intervals = [
                (timestamp - timestamps[0]).total_seconds()
                    for timestamp in timestamps[1:]
            ]
            frame_rates = [(i + 1) / interval for i, interval in enumerate(intervals)]
            return min(frame_rates) >= self._frame_rate

    def __nonzero__(self):
        if self.__bool__():
            return 1
        else:
            return 0


class VideoFrameDesc:
    """Descriptor that memorises the specs of a
    given video frame, without copying its data.
    """

    def __init__(self, frame):
        self.cols = frame.cols()
        self.rows = frame.rows()
        self.colour = frame.colour()
        self.data_length = frame.data_length()


class FileChecker(pgg.IObserver):
    """Descendant of GIFT-Grab's `Observer`, which
    listens to an `Observable` reading video frames
    from files and reports the file reader's specs,
    e.g. frame rate, frame count.
    """

    def __init__(self, file_reader):
        super(FileChecker, self).__init__()
        self._file_reader = file_reader
        self._frame_descs = []

    def attach(self):
        if self._file_reader is not None:
            self._file_reader.attach(self)

    def detach(self):
        if self._file_reader is not None:
            self._file_reader.detach(self)

    def update(self, frame):
        self._frame_descs.append(VideoFrameDesc(frame))

    def assert_data(self):
        return len(self._frame_descs) > 0

    def assert_colour(self, colour):
        for frame_desc in self._frame_descs:
            if frame_desc.colour != colour:
                return False
        return True

    def assert_frame_rate(self, frame_rate):
        if self._file_reader.get_frame_rate() != frame_rate:
            return False
        else:
            return True

    def assert_frame_dimensions(self,
                                frame_width, frame_height):
        for frame_desc in self._frame_descs:
            if frame_desc.cols != frame_width or\
               frame_desc.rows != frame_height:
                return False
        return True

    def assert_frame_data_lengths(self, colour,
                                  frame_width, frame_height):
        exp_data_length = pgg.VideoFrame.required_data_length(
            colour, frame_width, frame_height
            )

        for frame_desc in self._frame_descs:
            data_length = frame_desc.data_length
            if data_length != exp_data_length:
                return False
        return True



from pytest import yield_fixture, raises
from subprocess import check_call
from os import devnull, remove, listdir
from string import ascii_uppercase
from random import choice
from time import strftime
from pygiftgrab import Codec, VideoTargetFactory, VideoFrame, ColourSpace
from giftgrab.utils import inspection


# for easily removing created files when done
tmp_file_prefix = 'tmp_GiftGrab_test'
factory = None
target = None
frame = None
frame_with_colour_mismatch = None
file_name = None
frame_rate = 60


def __file_ext(codec):
    if codec == Codec.HEVC:
        return 'mp4'
    elif codec == Codec.Xvid:
        return 'avi'
    elif codec == Codec.VP9:
        return 'webm'


def __codec2str(codec):
    if codec == Codec.HEVC:
        return 'HEVC'
    elif codec == Codec.Xvid:
        return 'Xvid'
    elif codec == Codec.VP9:
        return 'VP9'


@yield_fixture(autouse=True)
def peri_test(codec, colour_space):
    # This section runs before each test

    global frame
    frame = VideoFrame(colour_space, 400, 640)

    global frame_with_colour_mismatch
    if colour_space == ColourSpace.BGRA:
        colour_space_with_mismatch = ColourSpace.I420
    else:
        colour_space_with_mismatch = ColourSpace.BGRA
    frame_with_colour_mismatch = VideoFrame(colour_space_with_mismatch,
                                            400, 640)

    global factory
    factory = VideoTargetFactory.get_instance()
    assert factory is not None

    global file_name
    file_name = '{}_{}-{}.{}'.format(
        tmp_file_prefix,
        strftime('%Y-%m-%d-%H-%M-%S'),
        ''.join(choice(ascii_uppercase) for _ in range(5)),
        __file_ext(codec)
    )

    global target, frame_rate
    target = None
    target = factory.create_file_writer(codec,
                                        file_name,
                                        frame_rate)
    assert target is not None

    try:
        file_null = open(devnull, 'w')
        check_call(['ffprobe', '--help'],
                   stdout=file_null, stderr=file_null)
    except BaseException:
        raise RuntimeError(
            'FFmpeg does not appear to be installed on your system,\n'
            'GiftGrab tests need FFmpeg.')

    # Run test
    yield

    # This section runs after each test

    global tmp_file_prefix
    for f in listdir('.'):
        if str(f).startswith(tmp_file_prefix):
            remove(f)


def test_append_with_colour_mismatch(codec):
    if codec != Codec.Xvid:
        return

    with raises(RuntimeError):
        target.append(frame_with_colour_mismatch)


def test_frame_rate():
    global target, file_name, frame_rate
    for i in range(10):
        target.append(frame)
    del target
    assert inspection.frame_rate(file_name) == frame_rate


def test_resolution(colour_space):
    rows = 1080
    cols = 1920
    frame1920x1080 = VideoFrame(colour_space, cols, rows)
    global target, file_name
    for i in range(10):
        target.append(frame1920x1080)
    del target
    assert inspection.resolution(file_name) == (cols, rows)


def test_num_frames():
    global target, file_name, frame, frame_rate
    num_frames = 3 * frame_rate
    for i in range(num_frames):
        target.append(frame)
    del target
    assert inspection.duration(file_name) * frame_rate == num_frames


def test_filetype_checked(codec):
    invalid_file_ext = 'rrr'
    assert not invalid_file_ext == __file_ext(codec)
    global factory, file_name
    target = None
    with raises(RuntimeError):
        target = factory.create_file_writer(
            codec,
            '{}.{}'.format(file_name, invalid_file_ext),
            frame_rate
        )
    assert target is None



#!/usr/bin/env python

from time import sleep, time, strftime
from datetime import timedelta
import yaml
from os.path import join, split, dirname, abspath
from os import makedirs
from random import choice
from string import ascii_uppercase
import logging
import pygiftgrab


MAX_X = 1920
MAX_Y = 1080


class Recorder:

    """A video recorder that can be started, paused, resumed, and stopped.

    When everything just works, this recorder will connect to either the DVI or SDI
    `port` of an Epiphan DVI2PCIe Duo, start grabbing frames there at the specified
    `frame_rate`, and save them using H265 (HEVC) encoding in an MP4 container
    pointed to by the specified `file_path`, which serves only as a prefix, i.e.
    for each resume-pause a new file will be created with a recording index and
    the file extension appended. After each video file has been finalised, a text
    file will be created as well that will contain the total duration for that
    video file.

    This recorder class will attempt to perform the requested operations within a
    `timeout_limit`. As such, after the timeout, the operations will fail and this
    will be propagated to the client through error messages. The recorder will
    simply stop at this point. Note that this recorder class does not attempt to be
    intelligent at all. As such, it is the client's responsibility to properly
    set the mentioned parameters.

    """

    def __init__(self, port, colour_space, frame_rate, file_path, timeout_limit=10):
        """Initialise the recorder with desired configuration.

        The recorder will not run if file writer cannot be created.

        @param port `pygiftgrab.Device.DVI2PCIeDuo_SDI` or
        `pygiftgrab.Device.DVI2PCIeDuo_DVI`
        @param colour_space `pygiftgrab.ColourSpace.BGRA` or
        `pygiftgrab.ColourSpace.I420`
        @param frame_rate Epiphan DVI2PCIe Duo supports up to 60 fps
        at ``1920 x 1080``, when using a single port, and 37 fps when using
        both ports (when capturing in the RGB24 colour space)
        @param file_path any referenced directories must exist and be
        writable by user. If a relative directory is specified, it is
        internally converted to an absolute one
        @param timeout_limit 5 attempts will be made for critical operations
        within this number of seconds (recorder will not start if this is
        negative)
        """

        self.timeout_limit = timeout_limit
        self.max_num_attempts = 5
        self.inter_attempt_duration = self.timeout_limit / self.max_num_attempts  # sec
        self.port = port
        self.colour_space = colour_space
        self.name = str(self.port)
        self.file = None
        self.is_running = False
        self.file_path = abspath(file_path)
        self.recording_index = 0
        self.frame_rate = frame_rate
        self.is_recording = False
        self.started_at = 0
        self.sub_frame = None
        self.device = None
        self.black_frame = None
        if self.timeout_limit > 0 \
           and (self.colour_space == pygiftgrab.ColourSpace.BGRA or \
                self.colour_space == pygiftgrab.ColourSpace.I420):
            self.is_running = True

    def start(self):
        """Connect to specified `port` and start data acquisition until
        `stop()`ped.

        Will simply do nothing if not `is_running`.
        """

        try:
            # mitigate re-start risk
            if not self.is_running:
                return

            if not self.__connect_device():
                return

            self.resume_recording()  # i.e. start recording

        except BaseException as e:
            logging.error('Unspecific exception on starting: ' + e.message +
                          ', aborting recording.')

    def stop(self):
        """Stop a `start()`ed recorder and disconnect from the
        Epiphan `port`.

        Will simply do nothing if not `is_running`.
        """
        if not self.is_running:
            return

        self.pause_recording()

        if self.__disconnect_device():
            self.is_running = False

    def pause_recording(self):
        """Pause a `start()`ed recorder and finalise the current
        video `file`.

        @see resume_recording()
        """
        # TODO - this line is quick hack because of GiftGrab#37
        if not self.is_running:
            return

        if not self.is_recording:
            return

        if self.__stop_acquisition():
            self.is_recording = False

            try:
                self.file.finalise()
                del self.file
                self.file = None
            except RuntimeError as e:
                logging.error(e.message)

            # write timing report as well
            try:
                report_file = open(self.__video_filename() + '.timing.yml', 'w')
                timing_report = dict(elapsed=str(timedelta(seconds=time() - self.started_at)))
                report_file.write(yaml.dump(timing_report, default_flow_style=False))
                report_file.close()
            except (IOError, yaml.YAMLError) as e:
                logging.error(
                    'Intermediate report generation failed with: ' +
                    e.message
                )


    def resume_recording(self):
        """Resume a `start()`ed and paused recorder.

        This will create and open a new video `file`.

        @see pause_recording()
        """
        # TODO - this line is quick hack because of GiftGrab#37
        if not self.is_running:
            return

        if self.is_recording:
            return

        if self.device:
            # TODO - rows and cols setting because of GiftGrab#40
            if self.sub_frame:
                self.device.set_sub_frame(self.sub_frame[0], self.sub_frame[1],
                                          self.sub_frame[2], self.sub_frame[3])
                cols = self.sub_frame[2]
                rows = self.sub_frame[3]
            else:
                self.device.get_full_frame()
                cols = MAX_X
                rows = MAX_Y

            tmp_frame = pygiftgrab.VideoFrame(self.colour_space, False)
            got_frame = self.device.get_frame(tmp_frame)
            if got_frame:
                cols = tmp_frame.cols()
                rows = tmp_frame.rows()
            self.black_frame = pygiftgrab.VideoFrame(self.colour_space, cols, rows)
        else:
            return

        if self.__create_video_writer():
            if self.__init_video_writer():
                if self.__start_acquisition():
                    self.started_at = time()
                    self.is_recording = True

    def set_sub_frame(self, x, y, width, height):
        """Set region of interest to record.

        This will have no effect if called in the middle of
        a recording.

        @param x upper-left corner coordinate
        @param y upper-left corner coordinate
        @param width
        @param height
        @throw ValueError if requested ROI values invalid,
        either due to negativity, or ouf-of-bounds (currently
        only up to ``1920 x 1080`` supported)
        """
        # TODO - hard-coded because of GiftGrab#42
        if not self.is_recording:
            if 0 <= x <= MAX_X and 0 <= y <= MAX_Y and \
               0 < width <= MAX_X and 0 < height <= MAX_Y and \
               x + width <= MAX_X and y + height <= MAX_Y:
                self.sub_frame = [x, y, width, height]
            else:
                raise ValueError('ROI ' + str(x) + ', ' + str(y) +
                                 ', ' + str(width) + ', ' + str(height) +
                                 ' invalid')

    def set_full_frame(self):
        """Tell recorder to use full resolution frames.

        This will have no effect if called in the middle of
        a recording.
        """
        if not self.is_recording:
            self.sub_frame = None

    def __start_acquisition(self):
        """Attach the current file writer to current device.

        @return ``True`` on success, ``False`` otherwise
        """
        try:
            self.device.attach(self.file)
        except RuntimeError as e:
            logging.error(
                'Could not start acquisition due to {}'.format(e.message)
            )
            return False
        else:
            return True

    def __stop_acquisition(self):
        """Detach the current file writer from current device.

        @return ``True`` on success, ``False`` otherwise
        """
        try:
            self.device.detach(self.file)
        except RuntimeError as e:
            logging.error(
                'Could not stop acquisition due to {}'.format(e.message)
            )
            return False
        else:
            return True

    def __video_filename(self):
        """Report what video file to record to.

        Automatically generate the full file path from
        `file_path` and `recording_index`.

        @return e.g. ``file_path-000003.mp4``
        """
        return self.file_path + '-' + \
            '{0:06d}'.format(self.recording_index) + \
            '.mp4'

    def __inter_frame_duration(self):
        """Compute time needed between frames based on `frame_rate`.

        @return
        """
        return 1.0 / self.frame_rate  # sec

    def __create_video_writer(self):
        """Attempt to create a writer for recording video files.

        @return: ``True`` if video writer created, ``False`` otherwise
        """
        attempts = 0
        while attempts < self.max_num_attempts:
            attempts += 1
            try:
                self.file = pygiftgrab.Factory.writer(pygiftgrab.Storage.File_HEVC)
            except RuntimeError as e:
                logging.error(
                    'Attempt #' + str(attempts) + ' of ' +
                    str(self.max_num_attempts) +
                    ' to get a video writer failed with: ' +
                    e.message
                )
                sleep(self.inter_attempt_duration)
                continue
            else:
                return True
        return False

    def __init_video_writer(self):
        """Attempt to initialise video writer with a new filename.

        @return ``True`` upon success, ``False`` otherwise
        """
        self.recording_index += 1
        filename = self.__video_filename()
        attempts = 0
        while attempts < self.max_num_attempts:
            attempts += 1
            try:
                self.file.init(filename, self.frame_rate)
            except RuntimeError as e:
                logging.error(
                    'Attempt #' + str(attempts) + ' of ' +
                    str(self.max_num_attempts) +
                    ' to initialise video writer failed with: ' +
                    e.message
                )
                sleep(self.inter_attempt_duration)
                continue
            else:
                return True
        return False

    def __connect_device(self):
        """Attempt to connect to device specified by `port`.

        @return: ``True`` if connected to `port`, ``False`` otherwise
        """
        attempts = 0
        while attempts < self.max_num_attempts:
            attempts += 1
            try:
                self.device = pygiftgrab.Factory.connect(self.port, self.colour_space)
            except IOError as e:
                logging.error(
                    'Attempt #' + str(attempts) + ' of ' +
                    str(self.max_num_attempts) +
                    ' to connect to ' + str(self.port) +
                    ' failed with: ' + e.message
                )
                sleep(self.inter_attempt_duration)
                continue
            else:
                return True
        return False

    def __disconnect_device(self):
        """Attempt to disconnect from device specified by `port`.

        @return: ``True`` if disconnected from `port`, ``False`` otherwise
        """
        attempts = 0
        while attempts < self.max_num_attempts:
            attempts += 1
            try:
                pygiftgrab.Factory.disconnect(self.port)
            except IOError as e:
                logging.error(
                    'Attempt #' + str(attempts) + ' of ' +
                    str(self.max_num_attempts) +
                    ' to disconnect from ' + str(self.port) +
                    ' failed with: ' + e.message
                )
                sleep(self.inter_attempt_duration)
                continue
            else:
                return True
        return False


def __port_to_str(epiphan_port):
    """Get string representation of given `epiphan_port`.

    @param epiphan_port
    @return
    @throw ValueError if invalid value given
    """
    if epiphan_port == pygiftgrab.Device.DVI2PCIeDuo_SDI:
        return 'SDI'
    elif epiphan_port == pygiftgrab.Device.DVI2PCIeDuo_DVI:
        return 'DVI'
    else:
        raise ValueError('Could not recognise port ' + epiphan_port)


def __str_to_port(epiphan_port):
    """Get actual port value from given `epiphan_port` string.

    @param epiphan_port
    @return
    @throw ValueError if invalid string given
    """
    if epiphan_port == 'SDI':
        return pygiftgrab.Device.DVI2PCIeDuo_SDI
    elif epiphan_port == 'DVI':
        return pygiftgrab.Device.DVI2PCIeDuo_DVI
    else:
        raise ValueError('Could not recognise port ' + epiphan_port)


def __colour_space_to_str(epiphan_colour_space):
    """Get string representation of given `epiphan_colour_space`.

    @param epiphan_colour_space
    @return
    @throw ValueError if invalid value given
    """
    if epiphan_colour_space == pygiftgrab.ColourSpace.BGRA:
        return 'BGRA'
    elif epiphan_colour_space == pygiftgrab.ColourSpace.I420:
        return 'I420'
    else:
        raise ValueError('Could not recognise colour space ' +
                         str(epiphan_colour_space))


def __str_to_colour_space(epiphan_colour_space):
    """Get actual colour space value from given `epiphan_colour_space` string.

    @param epiphan_colour_space
    @return
    @throw ValueError if invalid string given
    """
    if epiphan_colour_space == 'BGRA':
        return pygiftgrab.ColourSpace.BGRA
    elif epiphan_colour_space == 'I420':
        return pygiftgrab.ColourSpace.I420
    else:
        raise ValueError('Could not recognise colour space ' +
                         epiphan_colour_space)


def parse(file_path):
    """Parse `Recorder` configuration from given YAML file.

    Also create a session folder for this `Recorder`, using
    corresponding tag of given YAML file.

    @param file_path
    @return a ready-to-start `Recorder` on success
    @throw YAMLError if YAML file cannot be parsed
    @throw IOError if `file_path` cannot be opened for
    reading
    @throw ValueError if YAML file contains invalid values
    @throw OSError if unique session folder cannot be
    created
    """
    with open(file_path, 'r') as stream:
        # parse YAML file
        data = yaml.load(stream)

        # check everything specified properly
        if not data['file_path'] or not data['frame_rate'] \
           or not data['timeout_limit'] or not data['port'] \
           or not data['colour_space']:
            raise yaml.YAMLError('Could not parse ' + file_path)
        colour_space = __str_to_colour_space(data['colour_space'])
        frame_rate = data['frame_rate']
        if colour_space == pygiftgrab.ColourSpace.BGRA:
            max_frame_rate = 29
        elif colour_space == pygiftgrab.ColourSpace.I420:
            max_frame_rate = 59
        if not 0 < frame_rate <= max_frame_rate:
            raise ValueError('Supporting only up to ' + str(max_frame_rate) +
                             ' fps when using colour space ' +
                             data['colour_space'] + ' (' + str(frame_rate) + ' given)')
        timeout_limit = data['timeout_limit']
        if not 0 < timeout_limit <= 20:
            raise ValueError('Timeout should be positive, ' +
                             'and up to 20 sec.')
        port = __str_to_port(data['port'])
        data_file_path = abspath(data['file_path'])
        unique_file_path = None
        max_attempts = 5
        attempt = 0
        while True:
            attempt += 1
            try:
                unique_file_path = __session_folder(path_prefix=data_file_path)
            except OSError as e:
                if attempt >= max_attempts:
                    raise e
                else:
                    logging.error(e.message)
            else:
                break

        # create and return recorder
        return Recorder(port=port, colour_space=colour_space, frame_rate=frame_rate,
                        file_path=unique_file_path, timeout_limit=timeout_limit)


def dump(recorder):
    """Dump given `recorder`'s configuration to a YAML file.

    The file is deduced from `recorder` configuration.

    @param recorder
    @sa parse as same format is used
    @throw YAMLError if dumping fails
    @throw IOError if `file_path` cannot be opened for
    writing
    """
    file_path = join(dirname(recorder.file_path), 'config.yml')
    with open(file_path, 'w') as stream:
        port = __port_to_str(recorder.port)
        colour_space = __colour_space_to_str(recorder.colour_space)
        data = dict(port=port,
                    colour_space=colour_space,
                    frame_rate=recorder.frame_rate,
                    file_path=recorder.file_path,
                    timeout_limit=recorder.timeout_limit)
        stream.write(yaml.dump(data, default_flow_style=False))
        stream.close()


def __session_folder(path_prefix):
    """Attempt to create a session folder with a unique name.

    @param path_prefix last element treated as a file prefix,
    e.g. ``/my/dir/file``
    @return updated `path_prefix` after folder created
    @throw OSError if the attempt fails
    """
    path_prefix_folder, path_prefix_filename = split(path_prefix)
    unique_suffix = ''.join(choice(ascii_uppercase) for _ in range(5))
    attempted_folder = join(path_prefix_folder,
                            strftime('%Y-%m-%d-%H-%M-%S') +
                            '-' + unique_suffix)
    makedirs(attempted_folder)
    return join(attempted_folder, path_prefix_filename)



from pytest import fixture
from pygiftgrab import ColourSpace


def pytest_addoption(parser):
    parser.addoption('--colour-space', action='store',
                     help='Colour space specification (BGRA, I420, or UYVY)')
    parser.addoption('--frame-rate', action='store', type=int,
                     help='Frame-rate specification for test')
    parser.addoption('--observers', action='store', type=int,
                     help='Number of observers to use when testing frame rate')


@fixture(scope='session')
def frame_rate(request):
    return request.config.getoption('--frame-rate')


@fixture(scope='session')
def observers(request):
    return request.config.getoption('--observers')


@fixture(scope='session')
def colour_space(request):
    colour_space = request.config.getoption('--colour-space')
    case_insensitive = colour_space.lower()
    if case_insensitive == 'bgra':
        return ColourSpace.BGRA
    elif case_insensitive == 'i420':
        return ColourSpace.I420
    elif case_insensitive == 'uyvy':
        return ColourSpace.UYVY
    else:
        raise RuntimeError('Could not recognise colour space ' +
                           colour_space)




from subprocess import check_output


def frame_rate(file_path):
    """Use ``ffprobe`` to inspect frame rate of video file

    @param file_path
    @return
    @throw OSError if `file_path` does not exist, or if
    ``ffprobe`` cannot be run (e.g. not installed)
    @throw ValueError if frame rate cannot be inferred from
    ``ffprobe`` output
    """
    ffprobe_output = check_output(['ffprobe',
                                   '-v', 'error',
                                   '-of', 'flat=s=_',
                                   '-show_entries', 'stream=r_frame_rate',
                                   file_path])
    frame_rate_str = ffprobe_output.split('=')[1].strip()[1:-1].split('/')
    fr = -1
    if len(frame_rate_str) == 1:
        fr = float(frame_rate_str[0])
    elif len(frame_rate_str) == 2:
        fr = float(frame_rate_str[0])/float(frame_rate_str[1])
    else:
        raise ValueError('Frame rate cannot be inferred from ffprobe output')

    return fr


def duration(file_path):
    """Use ``ffprobe`` to inspect duration of video file

    @param file_path
    @return
    @throw OSError if `file_path` does not exist, or if
    ``ffprobe`` cannot be run (e.g. not installed)
    @throw ValueError if duration cannot be inferred from
    ``ffprobe`` output
    """
    ffprobe_output = check_output(['ffprobe',
                                   '-v', 'error',
                                   '-of', 'flat=s=_',
                                   '-show_entries', 'format=duration',
                                   file_path])
    duration_str = ffprobe_output.split('=')[1].strip()[1:-1]
    dur = -1
    if duration_str is not None:
        dur = float(duration_str)
    else:
        raise ValueError('Duration cannot be inferred from ffprobe output')

    return dur


def resolution(file_path):
    """Use ``ffprobe`` to inspect resolution of video file

    @param file_path
    @return width, height
    @throw OSError if `file_path` does not exist, or if
    ``ffprobe`` cannot be run (e.g. not installed)
    @throw ValueError if resolution cannot be inferred from
    ``ffprobe`` output
    """
    width = -1
    height = -1
    ffprobe_commands = [ ['ffprobe',
                          '-v', 'error',
                          '-of', 'flat=s=_',
                          '-show_entries', 'stream=width',
                          file_path],
                         ['ffprobe',
                          '-v', 'error',
                          '-of', 'flat=s=_',
                          '-show_entries', 'stream=height',
                          file_path] ]
    res = []
    for ffprobe_command in ffprobe_commands:
        ffprobe_output = check_output(ffprobe_command)
        resolution_str = ffprobe_output.split('=')[1].strip()
        if resolution_str is not None:
            res.append(float(resolution_str))
        else:
            raise ValueError('Resolution cannot be inferred from ffprobe output')
    width = res[0]
    height = res[1]

    return width, height


def codec(file_path):
    """Use ``ffprobe`` to inspect codec of video file

    @param file_path
    @return width, height
    @throw OSError if `file_path` does not exist, or if
    ``ffprobe`` cannot be run (e.g. not installed)
    @throw ValueError if codec cannot be inferred from
    ``ffprobe`` output
    """
    ffprobe_output = check_output(['ffprobe',
                                   '-v', 'error',
                                   '-of', 'flat=s=_',
                                   '-show_entries', 'stream=codec_name',
                                   file_path])
    codec_str = ffprobe_output.split('=')[1].strip().strip('"')
    if codec_str is not None:
        return codec_str
    else:
        raise ValueError('Codec cannot be inferred from ffprobe output')



from pytest import yield_fixture, fail, mark
from pygiftgrab import VideoSourceFactory, VideoFrame, ColourSpace

source = None
frame = None
frame_with_colour_mismatch = None
width = 0
height = 0
sub_x = 0
sub_y = 0
sub_width = 0
sub_height = 0
sub2_x = 0
sub2_y = 0
sub2_width = 0
sub2_height = 0


@yield_fixture(scope='session')
def peri_test(port, colour_space):
    # This section runs before each test
    factory = VideoSourceFactory.get_instance()
    global source
    try:
        source = factory.get_device(port, colour_space)
    except IOError as e:
        raise RuntimeError('Could not connect to Epiphan DVI2PCIe Duo,\n' +
                           'The detailed error was:\n' +
                           e.message)
    assert source is not None

    global frame, frame_with_colour_mismatch
    frame = VideoFrame(colour_space, False)
    if colour_space == ColourSpace.BGRA:
        mismatch_colour_space = ColourSpace.I420
    else:
        mismatch_colour_space = ColourSpace.BGRA
    frame_with_colour_mismatch = VideoFrame(mismatch_colour_space, False)

    global height, width
    assert source.get_frame(frame)
    width = frame.cols()
    height = frame.rows()

    global sub_x, sub_y, sub_width, sub_height
    sub_width = width // 3
    sub_height = height // 3
    sub_x = width // 2
    sub_y = height // 2
    # health checks
    assert sub_x + sub_width < width
    assert sub_y + sub_height < height

    global sub2_x, sub2_y, sub2_width, sub2_height
    sub2_width = width // 5
    sub2_height = height // 5
    sub2_x = width // 6
    sub2_y = height // 6
    # health checks
    assert sub2_x + sub2_width < width
    assert sub2_y + sub2_height < height

    # Run test
    yield

    # TODO: commented out to allow
    # for test GiftGrab#71 resolved
    # assert source is None


@mark.unit
@mark.usefixtures('peri_test')
def test_get_frame():
    assert source.get_frame(frame)
    assert frame.rows() > 0
    assert frame.cols() > 0

    assert not source.get_frame(frame_with_colour_mismatch)


@mark.unit
@mark.usefixtures('peri_test')
def test_sub_frame():
    source.set_sub_frame(sub_x, sub_y,
                         sub_width, sub_height)
    assert source.get_frame(frame)
    assert frame.cols() == sub_width
    assert frame.rows() == sub_height

    source.set_sub_frame(sub2_x, sub2_y,
                         sub2_width, sub2_height)
    assert source.get_frame(frame)
    assert frame.cols() == sub2_width
    assert frame.rows() == sub2_height


@mark.unit
@mark.usefixtures('peri_test')
def test_full_frame():
    source.set_sub_frame(sub_x, sub_y,
                         sub_width, sub_height)
    source.get_full_frame()
    assert source.get_frame(frame)
    assert frame.cols() == width
    assert frame.rows() == height



from pytest import mark, yield_fixture, fail, raises
from pygiftgrab import VideoFrame, ColourSpace
import numpy as np


frame = None
cols = 1920
rows = 1080


@yield_fixture(autouse=True)
def peri_test(colour_space):
    global frame
    frame = VideoFrame(colour_space, cols, rows)
    assert frame is not None

    yield


@mark.numpy_compatibility
def test_data():
    global frame, cols, rows
    data_np = frame.data()
    assert data_np.dtype == np.uint8
    assert frame.data_length() == data_np.size


@mark.numpy_compatibility
def test_data_length():
    global frame, cols, rows
    data_np = frame.data()
    data_len = frame.data_length()

    with raises(IndexError) as e:
        data_np[data_len]

    try:
        data_np[data_len - 1]
    except IndexError as e:
        fail(e.message)


@mark.numpy_compatibility
def test_read_access(colour_space):
    global frame
    global cols, rows
    data_np = frame.data()

    # general indices sampled at the boundaries only
    # intermediate indices should be accessible,
    # provided that these are accessible
    indices = [col * rows + row
               for col in [0, cols - 1]
               for row in [0, rows - 1]]

    # currently performing no value checks
    for index in indices:
        try:
            if colour_space == ColourSpace.BGRA:
                data_np[index * 4 + 0]  # Blue channel
                data_np[index * 4 + 1]  # Green channel
                data_np[index * 4 + 2]  # Red channel
                data_np[index * 4 + 3]  # Alpha channel

            elif colour_space == ColourSpace.I420:
                # see https://wiki.videolan.org/YUV/#I420
                data_np[index]         # Y plane
                data_np[(int)( index * 1.25 )]  # U plane
                data_np[(int)( index * 1.5 )]   # V plane

            elif colour_space == ColourSpace.UYVY:
                # see http://www.fourcc.org/pixel-format/yuv-uyvy/
                data_np[index * 2]
                data_np[index * 2 + 1]

        except IndexError as e:
            fail(e.message)


@mark.numpy_compatibility
def test_write_access():
    global frame
    global cols, rows
    data_np = frame.data()
    data_np_ = frame.data()  # shadow pointer
    data_len = frame.data_length()
    for i in [0, data_len // 4, data_len // 2, 3 * data_len // 4]:
        rolled_value = roll_uint8(data_np[i])
        assert rolled_value != data_np[i]
        data_np[i] = rolled_value
        assert data_np_[i] == rolled_value


def roll_uint8(value):
    """Roll given one-byte value, i.e. increment it such that
    the range is preserved.
    >>> roll_uint8(-10)
    ValueError: Acceptable range: 0 - 255
    >>> roll_uint8(-1)
    ValueError: Acceptable range: 0 - 255
    >>> roll_uint8(0)
    1
    >>> roll_uint8(10)
    11
    >>> roll_uint8(254)
    255
    >>> roll_uint8(255)
    0
    >>> roll_uint8(256)
    ValueError: Acceptable range: 0 - 255
    """
    new_value = value + 1
    if 0 < new_value < 256:
        return new_value
    elif new_value == 256:
        return 0
    else:
        raise ValueError('Acceptable range: 0 - 255')



#!/usr/bin/env python

from time import sleep, strptime
import pytest
import yaml
from subprocess import check_output
from os.path import isdir, dirname, isfile, join
from os import environ
from shutil import rmtree
from giftgrab.epiphan import parse, dump
import pygiftgrab
from giftgrab.utils.inspection import frame_rate, duration, resolution, codec


recorders = []


def timing_report(file_path):
    """Check whether `file_path` is a valid YAML timing report file

    @param file_path
    @return ``True`` if valid, ``False`` otherwise
    @throw IOError if `file_path` cannot be opened for reading
    @throw YAMLError if `file_path` cannot be parsed as a YAML file
    """
    with open(file_path, 'r') as stream:
        data = yaml.load(stream)
        if 'elapsed' not in data:
            return False
        try:
            strptime(data['elapsed'], '%H:%M:%S.%f')
        except ValueError:
            return False
    return True


@pytest.mark.epiphan_py_module
@pytest.mark.usefixtures('cleanup')
def test_parse(colour_space, config_dir):
    # not existing config file
    with pytest.raises(IOError):
        _ = parse(join('this', 'file', 'should', 'never', 'exist.yml'))

    # non-parseable file
    with pytest.raises(yaml.YAMLError):
        _ = parse(join(config_dir, 'yamlerror.yml'))

    # files with invalid values
    with pytest.raises(ValueError):
        _ = parse(join(config_dir, 'valueerror1.yml'))
    if colour_space == pygiftgrab.ColourSpace.BGRA:
        with pytest.raises(ValueError):
            _ = parse(join(config_dir, 'valueerror2.yml'))
    elif colour_space == pygiftgrab.ColourSpace.I420:
        with pytest.raises(ValueError):
            _ = parse(join(config_dir, 'valueerror3.yml'))
    else:
        pytest.fail('Colour space not configured properly')

    # folder that can't be created
    with pytest.raises(OSError):
        _ = parse(join(config_dir, 'oserror.yml'))


@pytest.mark.epiphan_py_module
@pytest.mark.usefixtures('cleanup')
def test_frame_grabbing(colour_space, config_dir):
    # test-input & data
    if colour_space == pygiftgrab.ColourSpace.BGRA:
        fs_config_file = join(config_dir, 'sdi.yml')
        us_config_file = join(config_dir, 'dvi.yml')
        fs_frame_rate = 28.0
        us_frame_rate = 14.0
    elif colour_space == pygiftgrab.ColourSpace.I420:
        fs_config_file = join(config_dir, 'sdi-i420.yml')
        us_config_file = join(config_dir, 'dvi-i420.yml')
        fs_frame_rate = 40.0
        us_frame_rate = 28.0
    else:
        pytest.fail('Colour space not configured properly')

    # create actual threads
    fs = parse(fs_config_file)
    us = parse(us_config_file)
    assert fs is not None
    recorders.append(fs)
    assert us is not None
    recorders.append(us)
    assert fs.frame_rate == fs_frame_rate and \
        fs.port == pygiftgrab.Device.DVI2PCIeDuo_SDI and \
        fs.timeout_limit == 20.0 and \
        fs.colour_space == colour_space
    assert us.frame_rate == us_frame_rate and \
        us.port == pygiftgrab.Device.DVI2PCIeDuo_DVI and \
        us.timeout_limit == 15.0 and \
        us.colour_space == colour_space

    # test actual output now
    roi = [426, 40, 1064, 1040]
    width_offset = 0
    # disabling the offset within #83
    if colour_space == pygiftgrab.ColourSpace.I420 and environ.get('USE_EPIPHANSDK'):
        width_offset = 4  # see GiftGrab issues #54 and #67
    fs.start()
    us.start()
    fs.set_sub_frame(roi[0], roi[1], roi[2], roi[3])  # should have no effect
    recording_duration = 5
    sleep(recording_duration+2)
    fs.pause_recording()
    us.pause_recording()
    sleep(4)

    # has output directory been created?
    assert isdir(dirname(fs.file_path))
    assert isdir(dirname(us.file_path))

    # have videos been recorded?
    fs_file_path = fs.file_path + '-000001.mp4'
    assert timing_report(fs_file_path + '.timing.yml')
    assert isfile(fs_file_path)
    us_file_path = us.file_path + '-000001.mp4'
    assert timing_report(us_file_path + '.timing.yml')
    assert isfile(us_file_path)
    assert frame_rate(fs_file_path) == fs.frame_rate
    assert frame_rate(us_file_path) == us.frame_rate
    assert duration(fs_file_path) >= recording_duration
    assert duration(us_file_path) >= recording_duration
    assert resolution(fs_file_path) == (1920, 1080)
    assert resolution(us_file_path) == (1920, 1080)
    assert codec(fs_file_path) == 'hevc'
    assert codec(us_file_path) == 'hevc'

    # another recording
    with pytest.raises(ValueError):
        fs.set_sub_frame(roi[0], -roi[1], roi[2], roi[3])
    fs.set_sub_frame(roi[0], roi[1], roi[2], roi[3])
    us.set_sub_frame(roi[0], roi[1], roi[2], roi[3])
    recording_duration = 10
    fs.resume_recording()
    us.resume_recording()
    sleep(recording_duration+2)
    fs.pause_recording()
    us.pause_recording()
    sleep(4)
    fs_file_path = fs.file_path + '-000002.mp4'
    assert timing_report(fs_file_path + '.timing.yml')
    assert isfile(fs_file_path)
    us_file_path = us.file_path + '-000002.mp4'
    assert isfile(us_file_path)
    assert timing_report(us_file_path + '.timing.yml')
    assert frame_rate(fs_file_path) == fs.frame_rate
    assert frame_rate(us_file_path) == us.frame_rate
    assert duration(fs_file_path) >= recording_duration
    assert duration(us_file_path) >= recording_duration
    assert resolution(fs_file_path) == (roi[2] - width_offset, roi[3])
    assert resolution(us_file_path) == (roi[2] - width_offset, roi[3])
    assert codec(fs_file_path) == 'hevc'
    assert codec(us_file_path) == 'hevc'

    # yet another recording
    recording_duration = 25
    fs.set_full_frame()
    fs.resume_recording()
    us.resume_recording()
    us.set_full_frame()  # should have no effect
    sleep(recording_duration+2)
    fs.pause_recording()
    us.pause_recording()
    sleep(4)
    fs_file_path = fs.file_path + '-000003.mp4'
    assert timing_report(fs_file_path + '.timing.yml')
    assert isfile(fs_file_path)
    us_file_path = us.file_path + '-000003.mp4'
    assert timing_report(us_file_path + '.timing.yml')
    assert isfile(us_file_path)
    assert frame_rate(fs_file_path) == fs.frame_rate
    assert frame_rate(us_file_path) == us.frame_rate
    assert duration(fs_file_path) >= recording_duration
    assert duration(us_file_path) >= recording_duration
    assert resolution(fs_file_path) == (1920, 1080)
    assert resolution(us_file_path) == (roi[2] - width_offset, roi[3])
    assert codec(fs_file_path) == 'hevc'
    assert codec(us_file_path) == 'hevc'

    # finish session
    fs.stop()
    us.stop()

    # test dumped Recorder configuration
    dump(fs)
    dumped_fs = parse(join(dirname(fs_file_path), 'config.yml'))
    assert dirname(dirname(dumped_fs.file_path)) == dirname(fs.file_path)
    assert dumped_fs.frame_rate == fs.frame_rate
    assert dumped_fs.port == fs.port
    assert dumped_fs.timeout_limit == fs.timeout_limit
    dump(us)
    dumped_us = parse(join(dirname(us_file_path), 'config.yml'))
    assert dirname(dirname(dumped_us.file_path)) == dirname(us.file_path)
    assert dumped_us.frame_rate == us.frame_rate
    assert dumped_us.port == us.port
    assert dumped_us.timeout_limit == us.timeout_limit


@pytest.yield_fixture
def cleanup():
    # This section runs before each test

    # Run test
    yield
    # This section runs after each test
    for recorder in recorders:
        recorder.stop()
    sleep(5)

    for recorder in recorders:
        rmtree(dirname(recorder.file_path))

    del recorders[:]



from pytest import yield_fixture, mark
from pygiftgrab import VideoSourceFactory, VideoFrame, ColourSpace
from time import sleep


source = None
frame = None
frame_with_colour_mismatch = None
tmp_frame = None


def generate_sub_frame(x_factor, y_factor, width_factor, height_factor):
    # full frame dimensions
    width, height = get_frame_dimensions()

    # generate sub frame dimensions
    sub_width = width // width_factor
    sub_height = height // height_factor
    sub_x = width // x_factor
    sub_y = height // y_factor

    # do health check on sub frame dimensions
    assert sub_x + sub_width < width
    assert sub_y + sub_height < height

    return sub_x, sub_y, sub_width, sub_height


def get_frame_dimensions():
    global source
    assert source is not None
    global tmp_frame
    assert source.get_frame(tmp_frame)
    return tmp_frame.cols(), tmp_frame.rows()


@yield_fixture(scope='session')
def peri_test(address, colour_space, init_delay):
    factory = VideoSourceFactory.get_instance()
    global source
    source = factory.connect_network_source(address, colour_space)
    sleep(init_delay)
    assert source is not None

    global frame, frame_with_colour_mismatch, tmp_frame
    frame = VideoFrame(colour_space, False)
    # TODO the loop is due to issue #135
    n_attempts = 0
    n_max_attempts = 10
    while not source.get_frame(frame):
        sleep(1)
        n_attempts += 1
        if n_attempts >= n_max_attempts:
            break
    assert source.get_frame(frame)
    tmp_frame = VideoFrame(colour_space, False)
    if colour_space == ColourSpace.BGRA:
        mismatch_colour_space = ColourSpace.I420
    else:
        mismatch_colour_space = ColourSpace.BGRA
    frame_with_colour_mismatch = VideoFrame(mismatch_colour_space, False)

    yield


@mark.unit
@mark.usefixtures('peri_test')
def test_get_frame():
    global source, frame
    assert source.get_frame(frame)
    assert frame.rows() > 0
    assert frame.cols() > 0

    assert not source.get_frame(frame_with_colour_mismatch)


@mark.unit
@mark.usefixtures('peri_test')
def test_full_frame():
    # ensure sub frame set before testing full frame
    width, height = get_frame_dimensions()
    sub_x, sub_y, sub_width, sub_height = generate_sub_frame(2, 2, 3, 3)
    global source
    source.set_sub_frame(sub_x, sub_y, sub_width, sub_height)

    # now test full frame
    source.get_full_frame()
    global frame
    assert source.get_frame(frame)
    assert frame.cols() == width
    assert frame.rows() == height



from pygiftgrab import Codec, ColourSpace


def pytest_addoption(parser):
    parser.addoption('--codec', action='store',
                     help='Codec (HEVC, Xvid, or VP9)')
    parser.addoption('--colour-space', action='store',
                     help='Colour space specification (BGRA, I420, or UYVY)')


def pytest_generate_tests(metafunc):
    if 'codec' in metafunc.fixturenames:
        codec_str = str(metafunc.config.option.codec)
        case_insensitive = codec_str.lower()
        if case_insensitive == 'hevc':
            codec = Codec.HEVC
        elif case_insensitive == 'xvid':
            codec = Codec.Xvid
        elif case_insensitive == 'vp9':
            codec = Codec.VP9
        else:
            raise RuntimeError('Could not recognise codec ' +
                               codec_str)
        metafunc.parametrize('codec', [codec])

    if 'colour_space' in metafunc.fixturenames:
        colour_space_str = str(metafunc.config.option.colour_space)
        case_insensitive = colour_space_str.lower()
        if case_insensitive == 'bgra':
            colour_space = ColourSpace.BGRA
        elif case_insensitive == 'i420':
            colour_space = ColourSpace.I420
        elif case_insensitive == 'uyvy':
            colour_space = ColourSpace.UYVY
        else:
            raise RuntimeError('Could not recognise colour space ' +
                               colour_space_str)
        metafunc.parametrize('colour_space', [colour_space])



from pytest import fixture


def pytest_addoption(parser):
    parser.addoption('--filepath', action='store', type=str, required=True,
                     help='Video file to use')


@fixture(scope='session')
def filepath(request):
    filepath = request.config.getoption('--filepath')
    if not filepath:
        raise RuntimeError('Filepath cannot be empty')
    else:
        return filepath



from pytest import fixture
from pygiftgrab import Codec, ColourSpace


def pytest_addoption(parser):
    parser.addoption('--colour-space', action='store', type=str, required=True,
                     help='Colour space specification (BGRA or I420)')
    parser.addoption('--filepath', action='store', type=str, required=True,
                     help='Video file to use')
    parser.addoption('--frame-rate', action='store', type=float, required=True,
                     help='Frame rate of video file')
    parser.addoption('--frame-count', action='store', type=int, required=True,
                     help='No. of frames in video file')
    parser.addoption('--frame-width', action='store', type=int, required=True,
                     help='Width of frames in video file')
    parser.addoption('--frame-height', action='store', type=int, required=True,
                     help='Height of frames in video file')


@fixture(scope='session')
def colour_space(request):
    colour_space = request.config.getoption('--colour-space')
    case_insensitive = colour_space.lower()
    if case_insensitive == 'bgra':
        return ColourSpace.BGRA
    elif case_insensitive == 'i420':
        return ColourSpace.I420
    elif case_insensitive == 'uyvy':
        return ColourSpace.UYVY
    else:
        raise RuntimeError('Could not recognise colour space ' +
                           colour_space)


@fixture(scope='session')
def filepath(request):
    filepath = request.config.getoption('--filepath')
    if not filepath:
        raise RuntimeError('Filepath cannot be empty')
    else:
        return filepath


@fixture(scope='session')
def frame_rate(request):
    frame_rate = float(request.config.getoption('--frame-rate'))
    if frame_rate <= 0:
        raise RuntimeError('Frame rate must be positive')
    else:
        return frame_rate


@fixture(scope='session')
def frame_count(request):
    frame_count = int(request.config.getoption('--frame-count'))
    if frame_count <= 0:
        raise RuntimeError('Frame count must be positive')
    else:
        return frame_count


@fixture(scope='session')
def frame_width(request):
    frame_width = int(request.config.getoption('--frame-width'))
    if frame_width <= 0:
        raise RuntimeError('Frame width must be positive')
    else:
        return frame_width


@fixture(scope='session')
def frame_height(request):
    frame_height = int(request.config.getoption('--frame-height'))
    if frame_height <= 0:
        raise RuntimeError('Frame height must be positive')
    else:
        return frame_height


from pytest import mark
try:
    # in case of PyPI installation, this will work:
    from giftgrab.tests.utils import FrameRateTimer
except ImportError:
    # in case of installation from source, this will work:
    from utils import FrameRateTimer
import pygiftgrab as pgg
import time


@mark.observer_pattern
def test_frame_rate(address, colour_space, frame_rate, observers, init_delay):
    timers = []
    for i in range(observers):
        timers.append(FrameRateTimer(frame_rate))
    factory = pgg.VideoSourceFactory.get_instance()
    source = factory.connect_network_source(address, colour_space)
    time.sleep(init_delay)

    for timer in timers:
        source.attach(timer)

    time.sleep(120)  # listen for data these nr. of sec
    for timer in timers:
        source.detach(timer)

    for timer in timers:
        assert timer



#!/usr/bin/env python
import pygiftgrab as gg
from time import strftime


def __unique_file_path():
    """Generate a timestamped, unique file name.

    @return
    """
    return strftime('%Y-%m-%d-%H-%M-%S')


def __extension(codec):
    """Infer file extension from passed `codec`.

    @param codec one of the codecs supported
    in pygiftgrab
    """
    if codec == gg.Codec.Xvid:
        return '.avi'
    elif codec == gg.Codec.HEVC:
        return '.mp4'
    elif codec == gg.Codec.VP9:
        return '.webm'
    else:
        raise ValueError('Codec unknown')


def __record_epiphan_dvi2pcieduo(port,
                                 colour_space,
                                 frame,
                                 num_frames,
                                 codec):
    """Record specified nr. of frames from specified Epiphan
    DVI2PCIeDuo port to specified file using specified codec.

    Note that the frame-rate of the file is set to 30, but
    this value may not reflect reality, due to the encoding
    time involved.

    @param port as defined in pygiftgrab
    @param frame a BGRA or I420 frame to use (i.e. determines
    the colour space)
    @param num_frames
    @param file_path does not include the file extension,
    which is automatically inferred using `codec`
    @param codec as defined in pygiftgrab
    """
    source_factory = gg.VideoSourceFactory.get_instance()
    epiphan = source_factory.get_device(port, colour_space)
    target_factory = gg.VideoTargetFactory.get_instance()
    frame_rate = 30
    filename = __unique_file_path() + __extension(codec)
    writer = target_factory.create_file_writer(codec,
                                               filename,
                                               frame_rate)
    for _ in range(num_frames):
        epiphan.get_frame(frame)
        writer.append(frame)


num_frames = 30


def record_epiphan_dvi2pcieduo_sdi_bgra_xvid():
    port = gg.Device.DVI2PCIeDuo_SDI
    colour_space = gg.ColourSpace.BGRA
    frame = gg.VideoFrame(colour_space, True)
    codec = gg.Codec.Xvid
    __record_epiphan_dvi2pcieduo(port,
                                 colour_space,
                                 frame,
                                 num_frames,
                                 codec)


def record_epiphan_dvi2pcieduo_sdi_bgra_hevc():
    port = gg.Device.DVI2PCIeDuo_SDI
    colour_space = gg.ColourSpace.BGRA
    frame = gg.VideoFrame(colour_space, True)
    codec = gg.Codec.HEVC
    __record_epiphan_dvi2pcieduo(port,
                                 colour_space,
                                 frame,
                                 num_frames,
                                 codec)


def record_epiphan_dvi2pcieduo_sdi_bgra_vp9():
    port = gg.Device.DVI2PCIeDuo_SDI
    colour_space = gg.ColourSpace.BGRA
    frame = gg.VideoFrame(colour_space, True)
    codec = gg.Codec.VP9
    __record_epiphan_dvi2pcieduo(port,
                                 colour_space,
                                 frame,
                                 num_frames,
                                 codec)


def record_epiphan_dvi2pcieduo_sdi_i420_xvid():
    port = gg.Device.DVI2PCIeDuo_SDI
    colour_space = gg.ColourSpace.I420
    frame = gg.VideoFrame(colour_space, True)
    codec = gg.Codec.Xvid
    __record_epiphan_dvi2pcieduo(port,
                                 colour_space,
                                 frame,
                                 num_frames,
                                 codec)


def record_epiphan_dvi2pcieduo_sdi_i420_hevc():
    port = gg.Device.DVI2PCIeDuo_SDI
    colour_space = gg.ColourSpace.I420
    frame = gg.VideoFrame(colour_space, True)
    codec = gg.Codec.HEVC
    __record_epiphan_dvi2pcieduo(port,
                                 colour_space,
                                 frame,
                                 num_frames,
                                 codec)


def record_epiphan_dvi2pcieduo_sdi_i420_vp9():
    port = gg.Device.DVI2PCIeDuo_SDI
    colour_space = gg.ColourSpace.I420
    frame = gg.VideoFrame(colour_space, True)
    codec = gg.Codec.VP9
    __record_epiphan_dvi2pcieduo(port,
                                 colour_space,
                                 frame,
                                 num_frames,
                                 codec)


def record_epiphan_dvi2pcieduo_dvi_bgra_xvid():
    port = gg.Device.DVI2PCIeDuo_DVI
    colour_space = gg.ColourSpace.BGRA
    frame = gg.VideoFrame(colour_space, True)
    codec = gg.Codec.Xvid
    __record_epiphan_dvi2pcieduo(port,
                                 colour_space,
                                 frame,
                                 num_frames,
                                 codec)


def record_epiphan_dvi2pcieduo_dvi_bgra_hevc():
    port = gg.Device.DVI2PCIeDuo_DVI
    colour_space = gg.ColourSpace.BGRA
    frame = gg.VideoFrame(colour_space, True)
    codec = gg.Codec.HEVC
    __record_epiphan_dvi2pcieduo(port,
                                 colour_space,
                                 frame,
                                 num_frames,
                                 codec)


def record_epiphan_dvi2pcieduo_dvi_bgra_vp9():
    port = gg.Device.DVI2PCIeDuo_DVI
    colour_space = gg.ColourSpace.BGRA
    frame = gg.VideoFrame(colour_space, True)
    codec = gg.Codec.VP9
    __record_epiphan_dvi2pcieduo(port,
                                 colour_space,
                                 frame,
                                 num_frames,
                                 codec)


def record_epiphan_dvi2pcieduo_dvi_i420_xvid():
    port = gg.Device.DVI2PCIeDuo_DVI
    colour_space = gg.ColourSpace.I420
    frame = gg.VideoFrame(colour_space, True)
    codec = gg.Codec.Xvid
    __record_epiphan_dvi2pcieduo(port,
                                 colour_space,
                                 frame,
                                 num_frames,
                                 codec)


def record_epiphan_dvi2pcieduo_dvi_i420_hevc():
    port = gg.Device.DVI2PCIeDuo_DVI
    colour_space = gg.ColourSpace.I420
    frame = gg.VideoFrame(colour_space, True)
    codec = gg.Codec.HEVC
    __record_epiphan_dvi2pcieduo(port,
                                 colour_space,
                                 frame,
                                 num_frames,
                                 codec)


def record_epiphan_dvi2pcieduo_dvi_i420_vp9():
    port = gg.Device.DVI2PCIeDuo_DVI
    colour_space = gg.ColourSpace.I420
    frame = gg.VideoFrame(colour_space, True)
    codec = gg.Codec.VP9
    __record_epiphan_dvi2pcieduo(port,
                                 colour_space,
                                 frame,
                                 num_frames,
                                 codec)

